@@ -1,7 +1,7 @@
 #ifndef SEQ_LOC_MAPPER_BASE__HPP
 #define SEQ_LOC_MAPPER_BASE__HPP
 
-/*  $Id: seq_loc_mapper_base.hpp 360367 2012-04-23 14:02:03Z grichenk $
+/*  $Id: seq_loc_mapper_base.hpp 572552 2018-10-16 15:50:18Z ivanov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
@@ -35,6 +35,7 @@
 
 #include <corelib/ncbistd.hpp>
 #include <corelib/ncbiobj.hpp>
+#include <corelib/ncbi_message.hpp>
 #include <util/range.hpp>
 #include <util/rangemap.hpp>
 #include <objects/seqloc/Na_strand.hpp>
@@ -67,7 +68,7 @@ class CSeq_graph;
 class IMapper_Sequence_Info;
 
 
-/// CMappingRange - describles a single interval to interval
+/// CMappingRange - describes a single interval to interval
 /// mapping.
 class NCBI_SEQ_EXPORT CMappingRange : public CObject
 {
@@ -81,9 +82,8 @@ public:
                   ENa_strand        dst_strand,
                   bool              ext_to = false,
                   int               frame = 0,
-                  TSeqPos           dst_total_len = kInvalidSeqPos,
                   TSeqPos           src_bioseq_len = kInvalidSeqPos,
-                  TSeqPos           dst_len = kInvalidSeqPos );
+                  TSeqPos           dst_len = kInvalidSeqPos);
 
     /// Check if the id is on the source sequence.
     bool GoodSrcId(const CSeq_id& id) const;
@@ -108,7 +108,7 @@ public:
     TRange Map_Range(TSeqPos           from,
                      TSeqPos           to,
                      const TRangeFuzz* fuzz = 0) const;
-    /// Map the strand, return true if the destination strand sould be
+    /// Map the strand, return true if the destination strand should be
     /// set (even if it's eNa_strand_unknown -- this may happen if the
     /// source strand is set to unknown).
     bool Map_Strand(bool is_set_strand,
@@ -129,11 +129,11 @@ private:
     CSeq_id_Handle      m_Dst_id_Handle;
     TSeqPos             m_Dst_from;
     ENa_strand          m_Dst_strand;
-    // Wether the mapping reverses the strand or not.
+    // Whether the mapping reverses the strand or not.
     // This value can be calculated from source and destination
     // strands, but is cached for better performance.
     bool                m_Reverse;
-    // Wether to extend the mapped location to the end of
+    // Whether to extend the mapped location to the end of
     // destination range. Used when mapping from a prot to a nuc.
     // ExtTo is set when both conditions are met:
     // - the mapping is from a protein to a nucleotide
@@ -144,9 +144,6 @@ private:
     bool                m_ExtTo;
     // Holds the frame shift (0 if none) of the underlying CDS (if any).
     int                 m_Frame;
-    // Holds the complete length of the destination.  This is needed
-    // to detect whether or not fuzzy edges should be extended to the end.
-    TSeqPos             m_Dst_total_len;
     // This holds the complete length of the original source bioseq.
     // Needed to detect whether or not fuzzy edges should be extended to the end.
     TSeqPos             m_Src_bioseq_len;
@@ -165,14 +162,15 @@ private:
     //friend class CSeq_align_Mapper;
     friend struct CMappingRangeRef_Less;
     friend struct CMappingRangeRef_LessRev;
+
 public:
     // Interface for CPairwiseAln converter
     TSeqPos GetSrc_from(void) const { return m_Src_from; }
     TSeqPos GetDst_from(void) const { return m_Dst_from; }
-    TSeqPos GetLength(void) const { return m_Src_to - m_Src_from; }
-    bool GetReverse(void) const { return m_Reverse; }
-    int GetGroup(void) const { return m_Group; }
-    void SetGroup(int grp) { m_Group = grp; }
+    TSeqPos GetLength(void)   const { return m_Src_to - m_Src_from; }
+    bool    GetReverse(void)  const { return m_Reverse; }
+    int     GetGroup(void)    const { return m_Group; }
+    void    SetGroup(int grp) { m_Group = grp; }
 };
 
 
@@ -274,6 +272,65 @@ private:
 };
 
 
+/////////////////////////////////////////////////////////////////////////////
+///
+///  CSeq_loc_Mapper_Options --
+///
+///  Options passed to CSeq_loc_Mapper[_Base] constructor.
+///
+
+class NCBI_SEQ_EXPORT CSeq_loc_Mapper_Options
+{
+public:
+    typedef int TMapOptions;
+
+    CSeq_loc_Mapper_Options(void);
+    CSeq_loc_Mapper_Options(IMapper_Sequence_Info* seq_info,
+                            TMapOptions opts = 0);
+    CSeq_loc_Mapper_Options(TMapOptions opts);
+
+    ///  Sequence type, length etc. provider. If any ids from the mapping
+    ///  ranges are not available through this object, they should be
+    ///  registered using CSeq_loc_Mapper_Base::SetSeqTypeById().
+    IMapper_Sequence_Info* GetMapperSequenceInfo(void) const;
+    CSeq_loc_Mapper_Options& SetMapperSequenceInfo(IMapper_Sequence_Info* seq_info);
+
+    /// Dense-seg mapping option.
+    /// @sa CSeq_loc_Mapper_Base::fAlign_Dense_seg_TotalRange
+    bool GetAlign_Dense_seg_TotalRange(void) const;
+    CSeq_loc_Mapper_Options& SetAlign_Dense_seg_TotalRange(bool value = true);
+
+    /// Mapping direction when mapping through a sparse-seg.
+    /// @sa CSeq_loc_Mapper_Base::fAlign_Sparse_ToFirst
+    /// @sa CSeq_loc_Mapper_Base::fAlign_Sparse_ToSecond
+    bool GetAlign_Sparse_ToFirst(void) const;
+    bool GetAlign_Sparse_ToSecond(void) const;
+    CSeq_loc_Mapper_Options& SetAlign_Sparse_ToFirst(bool value = true);
+    CSeq_loc_Mapper_Options& SetAlign_Sparse_ToSecond(bool value = true);
+
+    /// Mapping depth when using a seq-map, a bioseq or a GC-assembly.
+    /// @sa CSeq_loc_Mapper_Base::fMapSingleLevel
+    bool GetMapSingleLevel(void) const;
+    CSeq_loc_Mapper_Options& SetMapSingleLevel(bool value = true);
+
+    /// Mapped location trimming at sequence end. Off by default.
+    /// @sa CSeq_loc_Mapper_Base::fTrimMappedLocation
+    bool GetTrimMappedLocation(void) const;
+    CSeq_loc_Mapper_Options& SetTrimMappedLocation(bool value = true);
+
+private:
+    friend class CSeq_loc_Mapper_Base;
+
+    IMapper_Sequence_Info& GetSeqInfo(void) const;
+
+    bool x_IsSetOption(int opt) const;
+    void x_SetOption(int opt, bool enable);
+
+    mutable CRef<IMapper_Sequence_Info> m_SeqInfo;
+    TMapOptions m_Options;
+};
+
+
 /////////////////////////////////////////////////////////////////////////////
 ///
 ///  CSeq_loc_Mapper_Base --
@@ -281,7 +338,6 @@ private:
 ///  Mapping locations and alignments between bioseqs through seq-locs,
 ///  features, alignments or between parts of segmented bioseqs.
 
-
 class NCBI_SEQ_EXPORT CSeq_loc_Mapper_Base : public CObject
 {
 public:
@@ -295,11 +351,25 @@ public:
     enum EMapOptions {
         /// Ignore internal dense-seg structure - map each
         /// dense-seg according to the total ranges involved
-        fAlign_Dense_seg_TotalRange = 0x01,
+        fAlign_Dense_seg_TotalRange = 1 << 0,
+
         /// Flags used to indicate mapping direction when mapping
         /// through a sparse-seg.
-        fAlign_Sparse_ToFirst       = 0x00, ///< Map to first-id
-        fAlign_Sparse_ToSecond      = 0x02  ///< Map to second-id
+        fAlign_Sparse_ToFirst       = 0,      ///< Map to first-id
+        fAlign_Sparse_ToSecond      = 1 << 1, ///< Map to second-id
+        
+        /// Flag used when mapping through a seq-map (this includes
+        /// mapping through a bioseq or a GC-assembly). If set, each
+        /// call to Map() goes only one level up or down, unlike normal
+        /// mode which maps from any level as far up/down as possible.
+        /// The result of mapping can be mapped further by making another
+        /// call to Map().
+        fMapSingleLevel             = 1 << 2,
+
+        /// Enable trimming of source/destination ranges at sequence end.
+        /// By default locations can stretch beyond sequence end. With trimming
+        /// enabled the mapper will truncate ranges to fit sequence lengths.
+        fTrimMappedLocation         = 1 << 3
     };
     typedef int TMapOptions;
 
@@ -327,48 +397,62 @@ public:
     ///  NOTE: If the mapper is used with mixed sequence types, the
     ///  ranges must use genomic coordinates (for ranges on proteins
     ///  multiply all coordinates by 3).
-    /// @param seq_info
-    ///  Sequence type, length etc. provider. If any ids from the mapping
-    ///  ranges are not available through this object, they must be
-    ///  registered using SetSeqTypeById.
-    /// @sa IMapper_Sequence_Info
-    /// @sa SetSeqTypeById
-    CSeq_loc_Mapper_Base(CMappingRanges*        mapping_ranges,
-                         IMapper_Sequence_Info* seq_info = 0);
+    /// @param options
+    ///  Mapping options which need to be set during mapper initialization.
+    /// @sa CSeq_loc_Mapper_Options
+    CSeq_loc_Mapper_Base(CMappingRanges*         mapping_ranges,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
 
     /// Mapping through a feature, both location and product must be set.
-    CSeq_loc_Mapper_Base(const CSeq_feat&       map_feat,
-                         EFeatMapDirection      dir,
-                         IMapper_Sequence_Info* seq_info = 0);
+    CSeq_loc_Mapper_Base(const CSeq_feat&        map_feat,
+                         EFeatMapDirection       dir,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
 
     /// Mapping between two seq_locs.
-    CSeq_loc_Mapper_Base(const CSeq_loc&        source,
-                         const CSeq_loc&        target,
-                         IMapper_Sequence_Info* seq_info = 0);
+    CSeq_loc_Mapper_Base(const CSeq_loc&         source,
+                         const CSeq_loc&         target,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
 
     /// Mapping through an alignment. Need to specify target ID or
     /// target row of the alignment. Any other ID is mapped to the
     /// target one. Only the first row matching target ID is used,
     /// all other rows are considered source.
-    /// If the alignment is a spliced-seg, the merging is set to
-    /// 'merge by segment' automatically to merge mapped locations
-    /// by exon.
-    /// @sa SetMergeBySeg
+    CSeq_loc_Mapper_Base(const CSeq_align&       map_align,
+                         const CSeq_id&          to_id,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
+    /// Mapping through an alignment using specific source and target ids.
+    /// If the alignment is not one of dense-seg, dense-diag or packed-seg, the source
+    /// id is ignored.
+    CSeq_loc_Mapper_Base(const CSeq_id&          from_id,
+                         const CSeq_id&          to_id,
+                         const CSeq_align&       map_align,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
+    /// @deprecated Use the version with CSeq_loc_Mapper_Options instead.
+    NCBI_DEPRECATED
     CSeq_loc_Mapper_Base(const CSeq_align&      map_align,
                          const CSeq_id&         to_id,
-                         TMapOptions            opts = 0,
-                         IMapper_Sequence_Info* seq_info = 0);
+                         TMapOptions            opts,
+                         IMapper_Sequence_Info* seq_info);
+
     /// Sparse alignments require special row indexing since each
     /// row contains two seq-ids. Use options to specify mapping
     /// direction.
-    /// If the alignment is a spliced-seg, the merging is set to
-    /// 'merge by segment' automatically to merge mapped locations
-    /// by exon.
-    /// @sa SetMergeBySeg
+    CSeq_loc_Mapper_Base(const CSeq_align&       map_align,
+                         size_t                  to_row,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
+    /// Mapping through an alignment using specific source and target row numbers.
+    /// If the alignment is not one of dense-seg, dense-diag or packed-seg, the source
+    /// row is ignored.
+    CSeq_loc_Mapper_Base(size_t                  from_row,
+                         size_t                  to_row,
+                         const CSeq_align&       map_align,
+                         CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
+    /// @deprecated Use the version with CSeq_loc_Mapper_Options instead.
+    NCBI_DEPRECATED
     CSeq_loc_Mapper_Base(const CSeq_align&      map_align,
                          size_t                 to_row,
-                         TMapOptions            opts = 0,
-                         IMapper_Sequence_Info* seq_info = 0);
+                         TMapOptions            opts,
+                         IMapper_Sequence_Info* seq_info);
 
     ~CSeq_loc_Mapper_Base(void);
 
@@ -391,20 +475,24 @@ public:
     CSeq_loc_Mapper_Base& SetMergeAbutting(void);
     /// Merge only intervals from the same group. Group is created
     /// for each exon, dense-diag, std-seg and disc sub-alignment.
-    /// For spliced segs this mode is turned on by default.
     CSeq_loc_Mapper_Base& SetMergeBySeg(void);
     /// Merge intervals only if one is completely covered by another
     CSeq_loc_Mapper_Base& SetMergeContained(void);
     /// Merge any abutting or overlapping intervals
     CSeq_loc_Mapper_Base& SetMergeAll(void);
 
-    /// Wether to preserve or remove NULL sub-locations (usually
+    /// Whether to preserve or remove NULL sub-locations (usually
     /// indicating gaps) from the result. By default gaps are preserved.
     CSeq_loc_Mapper_Base& SetGapPreserve(void);
     CSeq_loc_Mapper_Base& SetGapRemove(void);
 
+    /// For mapping spliced-segs only: preserve or trim starting/ending
+    /// indels. By default indels are trimmed (only those at the whole
+    /// alignment start and end).
+    CSeq_loc_Mapper_Base& SetTrimSplicedSeg(bool trim);
+
     /// Keep ranges which can not be mapped. Does not affect truncation
-    /// of partially mapped ranges. By default nonmapping ranges are
+    /// of partially mapped ranges. By default non-mapping ranges are
     /// removed.
     CSeq_loc_Mapper_Base& KeepNonmappingRanges(void);
     CSeq_loc_Mapper_Base& TruncateNonmappingRanges(void);
@@ -414,6 +502,11 @@ public:
     /// correspond to the strand of the mapping source.
     CSeq_loc_Mapper_Base& SetCheckStrand(bool value = true);
 
+    /// When set to 'true' if mapped alignment has exactly one genomic and
+    /// one protein row, convert it to spliced-seg. By default all mixed-type
+    /// alignments are converted to std-seg.
+    CSeq_loc_Mapper_Base& MixedAlignsAsSpliced(bool value = true);
+
     /// Include source ranges in the mapped location. If turned
     /// on, the resulting seq-loc will be an equiv with the
     /// first sub-loc containing the usual mapped seq-loc, and
@@ -423,8 +516,15 @@ public:
     /// Merging mode must be set to MergeNone.
     CSeq_loc_Mapper_Base& IncludeSourceLocs(bool value = true);
 
+    /// Report source range trimming as an error. If the flag is set,
+    /// any trimming will result in throwing CAnnotMapperException.
+    /// Intended to be used when mapping GC-Assembly aliases.
+    CSeq_loc_Mapper_Base& SetErrorOnPartial(bool value = true);
+
     /// Map seq-loc
     CRef<CSeq_loc>   Map(const CSeq_loc& src_loc);
+    /// Take the total range from the location and run it through the mapper.
+    CRef<CSeq_loc>   MapTotalRange(const CSeq_loc& seq_loc);
     /// Map the whole alignment. Searches all rows for ranges
     /// which can be mapped.
     CRef<CSeq_align> Map(const CSeq_align& src_align);
@@ -435,22 +535,49 @@ public:
     /// The data may be truncated to match the new location.
     CRef<CSeq_graph> Map(const CSeq_graph& src_graph);
 
+    /// Flags defining seq-annot mapping options.
+    enum FAnnotMapFlag {
+        fAnnotMap_Location = 1 << 0, ///< Map seq-feat locations
+        fAnnotMap_Product  = 1 << 1, ///< Map seq-feat products
+        fAnnotMap_Both     = fAnnotMap_Location | fAnnotMap_Product,
+
+        /// Remove annotations which can not be mapped with this mapper.
+        /// If the flag is not set, the original annotation is stored
+        /// in the seq-annot.
+        fAnnotMap_RemoveNonMapping = 1 << 2,
+
+        /// Throw exception if an annotation can not be mapped.
+        fAnnotMap_ThrowOnFailure = 1 << 3,
+
+        fAnnotMap_Default = fAnnotMap_Both
+    };
+    typedef int TAnnotMapFlags;
+
+    /// Result of seq-annot mapping
+    enum EMapResult {
+        /// No annotation was mapped, the input seq-annot is unchanged.
+        eMapped_None = 0,
+        /// Some (not all) annotations were mapped.
+        eMapped_Some,
+        /// All annotations were mapped, none was removed.
+        eMapped_All
+    };
+
     /// Map each object from the Seq-annot and replace the original
     /// with the mapped one.
-    void Map(CSeq_annot& annot);
+    EMapResult Map(CSeq_annot& annot, TAnnotMapFlags flags = fAnnotMap_Default);
 
     /// Check if the last mapping resulted in partial location
     /// (not all ranges from the original location could be mapped
     /// to the target).
-    bool             LastIsPartial(void);
+    bool LastIsPartial(void);
 
-    typedef set<CSeq_id_Handle>        TSynonyms;
+    typedef set<CSeq_id_Handle> TSynonyms;
 
-    // Collect synonyms for the given seq-id and put them in the container.
-    // The default implementation just adds the id to the list of synonyms.
-    // Any overriden method should do at least the same.
-    void CollectSynonyms(const CSeq_id_Handle& id,
-                         TSynonyms&            synonyms) const;
+    // Collect synonyms for the id, store mapping of each synonym
+    // to the primary id. Returns primary id for the argument or the
+    // argument itself.
+    const CSeq_id_Handle& CollectSynonyms(const CSeq_id_Handle& id) const;
 
     // Sequence type - to recalculate coordinates.
     enum ESeqType {
@@ -488,17 +615,22 @@ protected:
     // only with cd-region features.
     void x_InitializeLocs(const CSeq_loc& source,
                           const CSeq_loc& target,
-                          int             frame = 0);
+                          int             src_frame = 0,
+                          int             dst_frame = 0);
     // Initialize the mapper from an alignment. Looks for the first
     // row containing the id and sets it as mapping target. All other
     // rows become mapping source.
     void x_InitializeAlign(const CSeq_align& map_align,
                            const CSeq_id&    to_id,
-                           TMapOptions       opts);
+                           const CSeq_id*    from_id = nullptr);
+    // Recursive version of the above.
+    void x_InitializeAlign(const CSeq_align& map_align,
+                           const TSynonyms&  to_ids,
+                           const TSynonyms*  from_ids = nullptr);
     // Initialize the mapper from an alignment, map to the specified row.
     void x_InitializeAlign(const CSeq_align& map_align,
                            size_t            to_row,
-                           TMapOptions       opts);
+                           size_t            from_row = size_t(-1));
 
     // Create dummy mapping from the whole destination location to itself.
     // This will prevent truncation of ranges already on the target.
@@ -523,7 +655,6 @@ protected:
                             const CInt_fuzz* fuzz_from = 0,
                             const CInt_fuzz* fuzz_to = 0,
                             int              frame = 0,
-                            TSeqPos          dst_total_len = kInvalidSeqPos,
                             TSeqPos          src_bioseq_len = kInvalidSeqPos);
 
     // Add new CMappingRange. This includes collecting all synonyms for the id,
@@ -538,7 +669,6 @@ protected:
                          TSeqPos        length,
                          bool           ext_right,
                          int            frame,
-                         TSeqPos        dst_total_len,
                          TSeqPos        src_bioseq_len,
                          TSeqPos        dst_length );
 
@@ -603,7 +733,7 @@ private:
         eMergeNone,      // no merging
         eMergeAbutting,  // merge only abutting intervals, keep overlapping
         eMergeContained, // merge if one range is contained in another
-        eMergeBySeg,     // merge ranges by mapping group (e.g. by exon)
+        eMergeBySeg,     // merge abutting and overlapping ranges by mapping group
         eMergeAll        // merge both abutting and overlapping intervals
     };
     enum EGapFlags {
@@ -618,12 +748,12 @@ private:
     // if the type can not be detected or there are multiple types.
     bool x_CheckSeqTypes(const CSeq_loc& loc,
                          ESeqType&       seqtype,
-                         TSeqPos&        len) const;
+                         TSeqPos&        len);
     // If x_CheckSeqTypes returns false, it may indicate that some
     // sequence types could not be detected. In this case the mapper
     // will attempt to find at least one known type in the location
     // and force it for all sub-locations with unknown types.
-    // The function will fail if there are diferent known types in the
+    // The function will fail if there are different known types in the
     // same seq-loc.
     ESeqType x_ForceSeqTypes(const CSeq_loc& loc) const;
 
@@ -639,16 +769,14 @@ private:
     TSeqPos x_GetRangeLength(const CSeq_loc_CI& it);
 
     // Initialize the mapper from different alignment types.
-    void x_InitAlign(const CDense_diag& diag, size_t to_row);
-    void x_InitAlign(const CDense_seg& denseg, size_t to_row,
-                     TMapOptions opts);
+    void x_InitAlign(const CDense_diag& diag, size_t to_row, size_t from_row);
+    void x_InitAlign(const CDense_seg& denseg, size_t to_row, size_t from_row);
     void x_InitAlign(const CStd_seg& sseg, size_t to_row);
-    void x_InitAlign(const CPacked_seg& pseg, size_t to_row);
+    void x_InitAlign(const CPacked_seg& pseg, size_t to_row, size_t from_row);
     void x_InitSpliced(const CSpliced_seg& spliced,
-                       const CSeq_id&      to_id);
+                       const TSynonyms&    to_ids);
     void x_InitSpliced(const CSpliced_seg& spliced, ESplicedRow to_row);
-    void x_InitSparse(const CSparse_seg& sparse, int to_row,
-                      TMapOptions opts);
+    void x_InitSparse(const CSparse_seg& sparse, int to_row);
 
     void x_IterateExonParts(const CSpliced_exon::TParts& parts,
                             ESplicedRow                  to_row,
@@ -696,6 +824,10 @@ private:
     // See also x_PushRangesToDstMix.
     void x_PushLocToDstMix(CRef<CSeq_loc> loc);
 
+    // Pushes NULL location to the destination mix (when a range
+    // can not be mapped).
+    void x_PushNullLoc(void);
+
     // Map the alignment. If row is NULL, map all rows. Otherwise
     // map only the selected row.
     CRef<CSeq_align> x_MapSeq_align(const CSeq_align& src_align,
@@ -762,17 +894,46 @@ private:
     void x_Map_PackedInt_Element(const CSeq_interval& si);
     void x_Map_PackedPnt_Element(const CPacked_seqpnt& pp, TSeqPos p);
 
+    // Get main seq-id for a synonym. If no mapping exists, returns the
+    // original id.
+    const CSeq_id_Handle& x_GetPrimaryId(const CSeq_id_Handle& synonym) const;
+
+    // Check if the id is in the list of synonyms.
+    bool x_IsSynonym(const CSeq_id& id, const TSynonyms& synonyms) const;
+
+    typedef map<CSeq_id_Handle, CSeq_id_Handle> TSynonymMap;
+    typedef map<CSeq_id_Handle, TSeqPos> TLengthMap;
+
+    friend class CTotalRangeSynonymMapper;
+
     // How to merge mapped locations.
     EMergeFlags          m_MergeFlag;
     // How to treat gaps (Null sub-locations) if any.
     EGapFlags            m_GapFlag;
-    // Wether to keep or discard ranges which can not be mapped.
-    bool                 m_KeepNonmapping;
-    // Wether to check or not if the original location is on the same strand
-    // as the mapping source.
-    bool                 m_CheckStrand;
-    // Wether to include a source of each mapped range to the mapped seq-loc.
-    bool                 m_IncludeSrcLocs;
+
+    // Other mapping options.
+    enum EMiscFlags {
+        // Trim leading/trailing indels (gaps) from mapped spliced-seg alignments.
+        fTrimSplicedSegs        = 1 << 0,
+        // Whether to keep or discard ranges which can not be mapped.
+        fKeepNonmapping         = 1 << 1,
+        // Whether to check or not if the original location is on the same strand
+        // as the mapping source.
+        fCheckStrand            = 1 << 2,
+        // Whether to include a source of each mapped range to the mapped seq-loc.
+        fIncludeSrcLocs         = 1 << 3,
+        // Prefer spliced-seg for mixed alignments.
+        fMixedAlignsAsSpliced   = 1 << 4,
+        // Treat any range truncation as an error (added for mapping to GC-Assembly
+        // aliases).
+        fErrorOnPartial         = 1 << 5
+    };
+    typedef int TMiscFlags;
+
+    bool x_IsSetMiscFlag(EMiscFlags flag) const { return (m_MiscFlags & flag) == flag; }
+    void x_SetMiscFlag(EMiscFlags flag, bool value);
+
+    TMiscFlags m_MiscFlags;
 
     // Mapped ranges collected from the currently parsed sub-location.
     mutable TRangesById  m_MappedLocs;
@@ -782,6 +943,12 @@ private:
     // Collected ranges for mapped graph. Used to adjust mapped graph data.
     CRef<CGraphRanges>   m_GraphRanges;
 
+    // Map each synonym to a primary seq-id.
+    mutable TSynonymMap  m_SynonymMap;
+
+    // Map each primary seq-id to sequence length.
+    mutable TLengthMap   m_LengthMap;
+
 protected:
     // Storage for sequence types.
     mutable TSeqTypeById m_SeqTypes;
@@ -801,13 +968,12 @@ protected:
     int                  m_CurrentGroup;
     // Control how fuzz is generated and propagated
     TFuzzOption          m_FuzzOption;
-
-    // Sequence info provider
-    mutable CRef<IMapper_Sequence_Info> m_SeqInfo;
+    // Misc mapping options
+    CSeq_loc_Mapper_Options m_MapOptions;
 
 public:
     // Initialize the mapper with default values
-    CSeq_loc_Mapper_Base(IMapper_Sequence_Info* seqinfo = 0);
+    CSeq_loc_Mapper_Base(CSeq_loc_Mapper_Options options = CSeq_loc_Mapper_Options());
 
     /// Methods for getting sequence types, use cached types (m_SeqTypes)
     /// if possible.
@@ -829,6 +995,17 @@ public:
 
     /// Get mapping ranges.
     const CMappingRanges& GetMappingRanges(void) const { return *m_Mappings; }
+
+    /// NOTE: In most cases CollectSynonyms(const CSeq_id_Handle& id) should
+    /// be used instead, since it takes care of synonym storage and mapping.
+    /// This method does nothing but storing synonyms in the container.
+    void CollectSynonyms(const CSeq_id_Handle& id, TSynonyms& synonyms) const;
+    // Check if ranges which can not be mapped should be replaced with NULL
+    // locations. By default removed ranges are reported using neighbor's fuzz.
+    // The flag is controlled from environment/registry:
+    // MAPPER_NONMAPPING_AS_NULL=t
+    // [Mapper]/Nonmapping_As_Null=t
+    static bool GetNonMappingAsNull(void);
 };
 
 
@@ -858,6 +1035,65 @@ public:
 };
 
 
+/////////////////////////////////////////////////////////////////////////////
+///
+///  CSeq_loc_Mapper_Message
+///
+///  Class used to report CSeq_loc_Mapper_Base issues through
+///  IMessageListener.
+class NCBI_SEQ_EXPORT CSeq_loc_Mapper_Message : public CMessage_Basic
+{
+public:
+    CSeq_loc_Mapper_Message(const string& msg,
+                            EDiagSev      sev,
+                            int           err_code = 0,
+                            int           sub_code = 0);
+    virtual ~CSeq_loc_Mapper_Message(void);
+
+    virtual CSeq_loc_Mapper_Message* Clone(void) const;
+    virtual void Write(CNcbiOstream& out) const;
+
+    enum EObjectType {
+        eNot_set,
+        eSeq_loc,
+        eSeq_feat,
+        eSeq_align,
+        eSeq_graph
+    };
+
+    /// Check type of the object stored in the message.
+    EObjectType Which(void) const { return m_ObjType; }
+
+    /// Set seq-loc object (copy into the message).
+    void SetLoc(const CSeq_loc& loc);
+    /// Get seq-loc object or null.
+    const CSeq_loc* GetLoc(void) const;
+
+    /// Set seq-feat object (copy into the message).
+    void SetFeat(const CSeq_feat& feat);
+    /// Get seq-feat object or null.
+    const CSeq_feat* GetFeat(void) const;
+
+    /// Set seq-align object (copy into the message).
+    void SetAlign(const CSeq_align& align);
+    /// Get seq-align object or null.
+    const CSeq_align* GetAlign(void) const;
+
+    /// Set seq-graph object (copy into the message).
+    void SetGraph(const CSeq_graph& graph);
+    /// Get seq-graph object or null.
+    const CSeq_graph* GetGraph(void) const;
+
+    /// Set the stored object to null.
+    void ResetObject(void);
+
+private:
+    EObjectType m_ObjType;
+
+    CRef<CObject> m_Obj;
+};
+
+
 struct CMappingRangeRef_Less
 {
     bool operator()(const CRef<CMappingRange>& x,
@@ -976,10 +1212,18 @@ CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::SetGapRemove(void)
 }
 
 
+inline
+CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::SetTrimSplicedSeg(bool trim)
+{
+    x_SetMiscFlag(fTrimSplicedSegs, trim);
+    return *this;
+}
+
+
 inline
 CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::SetCheckStrand(bool value)
 {
-    m_CheckStrand = value;
+    x_SetMiscFlag(fCheckStrand, value);
     return *this;
 }
 
@@ -994,7 +1238,7 @@ bool CSeq_loc_Mapper_Base::LastIsPartial(void)
 inline
 CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::KeepNonmappingRanges(void)
 {
-    m_KeepNonmapping = true;
+    x_SetMiscFlag(fKeepNonmapping, true);
     return *this;
 }
 
@@ -1002,7 +1246,15 @@ CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::KeepNonmappingRanges(void)
 inline
 CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::TruncateNonmappingRanges(void)
 {
-    m_KeepNonmapping = false;
+    x_SetMiscFlag(fKeepNonmapping, false);
+    return *this;
+}
+
+
+inline
+CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::MixedAlignsAsSpliced(bool value)
+{
+    x_SetMiscFlag(fMixedAlignsAsSpliced, value);
     return *this;
 }
 
@@ -1010,7 +1262,15 @@ CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::TruncateNonmappingRanges(void)
 inline
 CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::IncludeSourceLocs(bool value)
 {
-    m_IncludeSrcLocs = value;
+    x_SetMiscFlag(fIncludeSrcLocs, value);
+    return *this;
+}
+
+
+inline
+CSeq_loc_Mapper_Base& CSeq_loc_Mapper_Base::SetErrorOnPartial(bool value)
+{
+    x_SetMiscFlag(fErrorOnPartial, value);
     return *this;
 }
 
@@ -1034,11 +1294,12 @@ inline
 CSeq_loc_Mapper_Base::ESeqType
 CSeq_loc_Mapper_Base::GetSeqTypeById(const CSeq_id_Handle& idh) const
 {
-    TSeqTypeById::const_iterator it = m_SeqTypes.find(idh);
+    CSeq_id_Handle primary_id = CollectSynonyms(idh);
+    TSeqTypeById::const_iterator it = m_SeqTypes.find(primary_id);
     if (it != m_SeqTypes.end()) {
         return it->second;
     }
-    return GetSeqType(idh);
+    return GetSeqType(primary_id);
 }
 
 
@@ -1072,6 +1333,122 @@ int CSeq_loc_Mapper_Base::GetWidthById(const CSeq_id& id) const
 }
 
 
+inline
+CSeq_loc_Mapper_Options::CSeq_loc_Mapper_Options(void)
+    : m_SeqInfo(0), m_Options(0) {}
+
+inline
+CSeq_loc_Mapper_Options::CSeq_loc_Mapper_Options(IMapper_Sequence_Info* seq_info,
+                                                 TMapOptions            opts)
+    : m_SeqInfo(seq_info), m_Options(opts) {}
+
+inline
+CSeq_loc_Mapper_Options::CSeq_loc_Mapper_Options(TMapOptions opts)
+    : m_SeqInfo(0), m_Options(opts) {}
+
+inline
+IMapper_Sequence_Info*
+CSeq_loc_Mapper_Options::GetMapperSequenceInfo(void) const
+{
+    return m_SeqInfo;
+}
+
+inline
+CSeq_loc_Mapper_Options&
+CSeq_loc_Mapper_Options::SetMapperSequenceInfo(IMapper_Sequence_Info* seq_info)
+{
+    m_SeqInfo = seq_info;
+    return *this;
+}
+
+inline
+bool CSeq_loc_Mapper_Options::GetAlign_Dense_seg_TotalRange(void) const
+{
+    return x_IsSetOption(CSeq_loc_Mapper_Base::fAlign_Dense_seg_TotalRange);
+}
+
+inline
+CSeq_loc_Mapper_Options&
+CSeq_loc_Mapper_Options::SetAlign_Dense_seg_TotalRange(bool value)
+{
+    x_SetOption(CSeq_loc_Mapper_Base::fAlign_Dense_seg_TotalRange, value);
+    return *this;
+}
+
+inline
+bool CSeq_loc_Mapper_Options::GetAlign_Sparse_ToFirst(void) const
+{
+    return !x_IsSetOption(CSeq_loc_Mapper_Base::fAlign_Sparse_ToSecond);
+}
+
+inline
+bool CSeq_loc_Mapper_Options::GetAlign_Sparse_ToSecond(void) const
+{
+    return x_IsSetOption(CSeq_loc_Mapper_Base::fAlign_Sparse_ToSecond);
+}
+
+inline
+CSeq_loc_Mapper_Options&
+CSeq_loc_Mapper_Options::SetAlign_Sparse_ToFirst(bool value)
+{
+    x_SetOption(CSeq_loc_Mapper_Base::fAlign_Sparse_ToSecond, !value);
+    return *this;
+}
+
+inline
+CSeq_loc_Mapper_Options&
+CSeq_loc_Mapper_Options::SetAlign_Sparse_ToSecond(bool value)
+{
+    x_SetOption(CSeq_loc_Mapper_Base::fAlign_Sparse_ToSecond, value);
+    return *this;
+}
+
+inline
+bool CSeq_loc_Mapper_Options::GetMapSingleLevel(void) const
+{
+    return x_IsSetOption(CSeq_loc_Mapper_Base::fMapSingleLevel);
+}
+
+inline
+CSeq_loc_Mapper_Options&
+CSeq_loc_Mapper_Options::SetMapSingleLevel(bool value)
+{
+    x_SetOption(CSeq_loc_Mapper_Base::fMapSingleLevel, value);
+    return *this;
+}
+
+inline
+bool CSeq_loc_Mapper_Options::GetTrimMappedLocation(void) const
+{
+    return x_IsSetOption(CSeq_loc_Mapper_Base::fTrimMappedLocation);
+}
+
+inline
+CSeq_loc_Mapper_Options&
+CSeq_loc_Mapper_Options::SetTrimMappedLocation(bool value)
+{
+    x_SetOption(CSeq_loc_Mapper_Base::fTrimMappedLocation, value);
+    return *this;
+}
+
+inline
+bool CSeq_loc_Mapper_Options::x_IsSetOption(int opt) const
+{
+    return (m_Options & opt) != 0;
+}
+
+inline
+void CSeq_loc_Mapper_Options::x_SetOption(int opt, bool enable)
+{
+    if ( enable ) {
+        m_Options |= opt;
+    }
+    else {
+        m_Options &= ~opt;
+    }
+}
+
+
 /* @} */
 
 
